#include "render_text.h"

#include <GL/glew.h>
#include <stdio.h>
#include <string.h>

#include "utils_graphics.h"
#include "render_text_cp1250.h"
#include "shaders.h"

#define MAX_BUF     8912
#define POSITION    0


char *text_vertex_shader_glsl = "#version 330\n" \
    "uniform mat4 u_mvp_matrix;\n" \
    "uniform vec2 u_pos;\n" \
    "#define POSITION 0\n" \
    "layout( location = POSITION ) in vec2 in_position;\n" \
    "void main()\n" \
    "{\n" \
    "    gl_Position = u_mvp_matrix * vec4(in_position + u_pos, 0.0, 1.0 );\n" \
    "}\n";

char *text_fragment_shader_glsl = "#version 150\n" \
    "uniform vec4 u_color;\n" \
    "out vec4 out_color;\n" \
    "void main()\n" \
    "{\n" \
    "    out_color = u_color;\n" \
    "}\n";

GLuint text_vertex_shader;
GLuint text_fragment_shader;
GLuint text_vertex_array;
GLuint text_vertex_buffer;
GLuint text_program;


// położenie danych współrzędnych punktów poszczególnych znaków
// w buforze; dane znaków zaczynają się od spacji
static int text_positions[224];

// liczba danych współrzędnych punktów poszczególnych znaków
// w buforze; dane znaków zaczynają się od spacji
static int text_sizes[224];

//////////////////////////////////////////////////////////////////////
// inicjacja mechanizmów używanych podczas renderingu tekstu
//////////////////////////////////////////////////////////////////////
unsigned char init_draw_text() {
    // bufor współrzędnych punktów
    GLfloat buf[MAX_BUF];
    // położenie bieżących danych w buforze współrzędnych punktów
    int buf_pos;
    // licznik pętli
    unsigned int i;
    int col, row;
    unsigned char chr;


    // wczytanie i przygotowanie obsługi programu rysującego napisy
    text_program = glCreateProgram();
    if (!text_program) {
        fprintf(stderr, "Obiekt programu nie jest zainicjalizowany\n");
        return 0;
    }

    text_vertex_shader = load_shader_from_memory(GL_VERTEX_SHADER, text_vertex_shader_glsl);
    if (!text_vertex_shader) {
        return 0;
    }
    glAttachShader(text_program, text_vertex_shader);

    text_fragment_shader = load_shader_from_memory(GL_FRAGMENT_SHADER, text_fragment_shader_glsl);
    if (!text_fragment_shader) {
        return 0;
    }
    glAttachShader(text_program, text_fragment_shader);
    if (!link_and_validate_program(text_program)) {
        return 0;
    }

    // obiekt tablic wierzchołków z buforem tablic wierzchołków
    // na współrzędne punktów z tekstem (bez ładowania danych)
    buf_pos = 0;
    // generowanie poszczególnych liter, począwszy od spacji
    for(i = 0; i < 224 && buf_pos < MAX_BUF; i++) {
        // położenie danych w buforze; wartość dzielimy przez 2,
        // bowiem położenie każdego wierzchołka opisują dwie współrzędne
        text_positions[i] = buf_pos / 2;

        // przygotowanie danych pojedynczego znaku
        for (row = 0; row < 16; row++) {
            // pobranie danych pojedynczego wiersza pikseli w znaku
            // (dane znaków zaczynają się od spacji)
            chr = font8x16CP1250[i][row];
            for (col = 0; col < 8; col++) {
                // test obecności poszczególnych bitów oraz test pojemości bufora
                if (((chr >> col) & 0x01) == 0x01 && buf_pos < MAX_BUF - 2) {
                    buf[buf_pos + 0] = (GLfloat)(7 - col);
                    buf[buf_pos + 1] = (GLfloat)row;
                    buf_pos += 2;
                }
            }
        }

        // rozmiar danych w buforze; wartość dzielimy przez 2,
        // bowiem położenie każdego wierzchołka opisują dwie współrzędne
        text_sizes[i] = buf_pos/2 - text_positions[i];
    }

/*
        for(i = 0; i < 224; i++) {
            printf("text_positions[%d]: %d,  text_sizes[%d]: %d\n", i, text_positions[i], i, text_sizes[i]);
        }
*/

    glGenVertexArrays(1, &text_vertex_array);
    glBindVertexArray(text_vertex_array);

    glGenBuffers(1, &text_vertex_buffer);
    glBindBuffer(GL_ARRAY_BUFFER, text_vertex_buffer);
    glBufferData(GL_ARRAY_BUFFER, buf_pos * sizeof(GLfloat), buf, GL_STATIC_DRAW);

    // wyłączenie obiektu tablic wierzchołków
    glBindVertexArray(0);

    //printf("vertex_array: %d,  vertex_buffer: %d,  buf_pos: %d\n", text_vertex_array, text_vertex_buffer, buf_pos);
/*
        for (i = 0; i < buf_pos; i++) {
            printf("buf[%4d]: %.2f\n", i, buf[i]);
        }
*/
    return 1;
}

// funkcja rysująca napis w wybranym miejscu
// x, y - współrzędne okienkowe początku napisu
// str - tekst napisu
void draw_text(float x, float y, char *str, float size, float r, float g, float b, float alpha) {
    float projection_matrix[16], aux_matrix[16];
    float color[4];
    GLboolean program_point_size;
    size_t str_len;
    unsigned char ch;
    unsigned int i;
    GLint viewport[4], width, height;
    float unit_size = 400.0f;
    float left = 0.0f;
    float right = unit_size;
    float bottom = 0.0f;
    float top = unit_size;
    float near = 0.0f;
    float far = unit_size;
    float point_size;

    if (!str) {
        return;
    }
    str_len = strlen(str);
    if (str_len < 1) {
        return;
    }

    //printf("%s,  len: %d\n", str, (int)str_len);

    color[0] = r;
    color[1] = g;
    color[2] = b;
    color[3] = alpha;

    unit_size = 14.0f / size * 400.0f;
    right = unit_size;
    top = unit_size;
    far = unit_size;


    // pobranie rozmiarów okna renderingu
    glGetIntegerv(GL_VIEWPORT, viewport);
    width = viewport[2];
    height = viewport[3];
    //printf("[0]: %d, [1]: %d, [2]: %d, [3]: %d\n", viewport[0], viewport[1], viewport[2], viewport[3]);
    if (width < height && width > 0) {
        get_ortho_matrix(projection_matrix, left, right, bottom*height/width, top*height/width, near, far);
    } else if (width >= height && height > 0) {
        get_ortho_matrix(projection_matrix, left*width/height, right*width/height, bottom, top, near, far);
    } else {
        get_ortho_matrix(projection_matrix, left, right, bottom, top, near, far);
    }

    // przygotowanie macierzy rzutowania prostokątnego 2D; wykorzystujemy
    // wyłącznie macierz rzutowania, bowiem macierz modelu-widoku jest
    // w tym przypadku macierzą jednostkową
    //get_ortho_matrix(projection_matrix, 0.0f, (float)viewport[2], 0.0f, (float)viewport[3], -2.0f, 2.0f);

    //dump_matrix4x4(projection_matrix);

    // sprawdzenie czy włączone jest generowanie wielkości punktu w shaderze
    program_point_size = glIsEnabled(GL_PROGRAM_POINT_SIZE);
    if (program_point_size) {
        glDisable(GL_PROGRAM_POINT_SIZE);
    }

    // rozmiar punktów
    if (size < 11.9f) {
        point_size = 1.2f;
    } else if (size < 14.0f) {
        point_size = 1.45f;
    } else if (size < 16.0f) {
        point_size = 1.5f;
    } else {
        point_size = 1.5f;
    }
    glPointSize(point_size);

    //printf("%d, %d, %d\n", text_vertex_array, text_vertex_buffer, text_program);

    // włączenie obiektu tablic wierzchołków
    glBindVertexArray(text_vertex_array);
    glBindBuffer(GL_ARRAY_BUFFER, text_vertex_buffer);
    glVertexAttribPointer(POSITION, 2, GL_FLOAT, GL_FALSE, 0, NULL);
    glEnableVertexAttribArray(POSITION);

    // włączenie programu
    glUseProgram(text_program);

    // załadowanie zmiennych jednorodnych - iloczynu macierzy
    // modelu-widoku i rzutowania oraz koloru punktu
    // macierz model_view jest tu macierzą ednostkową, więc jest pomijana w obliczeniach
    transpose_matrix(aux_matrix, projection_matrix, 4);
    glUniformMatrix4fv(glGetUniformLocation(text_program, "u_mvp_matrix" ), 1, GL_FALSE, aux_matrix);
    glUniform4fv(glGetUniformLocation(text_program, "u_color" ), 1, color);

    // rysowanie poszczególnych liter
    for (i = 0; i < str_len; i++) {
        // pobranie znaku
        ch = (unsigned char)str[i];

        // pomijamy spacje i znaki sterujące
        if (ch > 32) {
            // położenie znaku
            glUniform2f(glGetUniformLocation(text_program, "u_pos"), (GLfloat)(x + i * 10), (GLfloat)y);
            // narysowanie danych zawartych w tablicach wierzchołków
            //printf("x: %.2f, y: %.2f, '%c', %d, %d, %d\n", x, y, ch, ch, text_positions[ch-32], text_sizes[ch-32]);
            glDrawArrays(GL_POINTS, text_positions[ch-32], text_sizes[ch-32]);
        }
    }

    // wyłączenie programu
    glUseProgram(0);

    // wyłączenie obiektu tablic wierzchołków
    glBindVertexArray(0);

    // w razie potrzeby włączamy generowanie wielkości punktu w shaderze
    if (program_point_size) {
        glEnable(GL_PROGRAM_POINT_SIZE);
    }
}

// usunięcie mechanizmów używanych podczas renderingu tekstu
void cleanup_draw_text() {
    //printf("cleanup_draw_text\n");
    if (text_program && text_fragment_shader) {
        glDetachShader(text_program, text_fragment_shader);
    }
    if (text_program && text_vertex_shader) {
        glDetachShader(text_program, text_vertex_shader);
    }
    if (text_fragment_shader) {
        glDeleteShader(text_fragment_shader);
    }
    if (text_vertex_shader) {
        glDeleteShader(text_vertex_shader);
    }
    if (text_program) {
        glDeleteProgram(text_program);
    }

    glDeleteBuffers(1, &text_vertex_buffer);
    if (text_vertex_array) {
        glDeleteVertexArrays(1, &text_vertex_array);
    }

}
