#include "shaders.h"

#include <stdlib.h>
#include <stdio.h>

char *text_file_read(char *fname);

GLuint load_shader(const GLenum type, char *filename) {
    GLuint shader;
    GLint status;
    GLchar *shader_source = text_file_read(filename);

    if (!shader_source) {
        printf("Couldn't read shader source: %s\n", filename);
        return 0;
    }
    //printf("%s\n---\n", shader_source);

    shader = glCreateShader(type);
    if (!shader) {
        printf("Błąd podczas tworzenia shadera: %s [%d]\n", filename, shader);
        return 0;
    }
    //printf("shader: %d\n", shader);
    glShaderSource(shader, 1, (const char* const*)&shader_source, NULL);
    free(shader_source);

    glCompileShader(shader);
    glGetShaderiv(shader, GL_COMPILE_STATUS, &status);
    if (status != GL_TRUE) {
        printf("Niepoprawna kompilacja shadera: %s [%d]\n", filename, shader);
        print_shader_info_log(shader);
        if (shader) {
            glDeleteShader(shader);
        }
        return 0;
    }

    return shader;
}

GLuint load_shader_from_memory(const GLenum type, GLchar *shader_source) {
    GLuint shader;
    GLint status;

    if (!shader_source) {
        printf("Couldn't read shader source: %s\n", shader_source);
        return 0;
    }
    //printf("%s\n---\n", shader_source);

    shader = glCreateShader(type);
    if (!shader) {
        printf("Błąd podczas tworzenia shadera: %s [%d]\n", shader_source, shader);
        return 0;
    }
    //printf("shader: %d\n", shader);
    glShaderSource(shader, 1, (const char* const*)&shader_source, NULL);

    glCompileShader(shader);
    glGetShaderiv(shader, GL_COMPILE_STATUS, &status);
    if (status != GL_TRUE) {
        printf("Niepoprawna kompilacja shadera: %s [%d]\n", shader_source, shader);
        print_shader_info_log(shader);
        if (shader) {
            glDeleteShader(shader);
        }
        return 0;
    }

    return shader;
}


GLuint link_and_validate_program(GLuint program) {
    GLint status;

    // konsolidacja shaderów
    glLinkProgram(program);
    glGetProgramiv(program, GL_LINK_STATUS, &status);
    if (status == GL_FALSE) {
        printf("Niepoprawna konsolidacja programu\n");
        print_program_info_log(program);
        return 0;
    }

    glValidateProgram(program);
    glGetProgramiv(program, GL_VALIDATE_STATUS, &status);
    if (status == GL_FALSE) {
        printf("Niepoprawna walidacja programu\n");
        print_program_info_log(program);
        return 0;
    }

    return 1;
}


void print_shader_info_log(GLuint obj) {
    GLint infologLength = 0;
    int charsWritten  = 0;
    char *infoLog;

    glGetShaderiv(obj, GL_INFO_LOG_LENGTH, &infologLength);

    if (infologLength > 0) {
        infoLog = (char*)malloc(infologLength);
        glGetShaderInfoLog(obj, infologLength, &charsWritten, infoLog);
        printf("%s\n", infoLog);
        free(infoLog);
    }
}


void print_program_info_log(GLuint obj) {
    int infologLength = 0;
    int charsWritten  = 0;
    char *infoLog;

    glGetProgramiv(obj, GL_INFO_LOG_LENGTH, &infologLength);

    if (infologLength > 0) {
        infoLog = (char*)malloc(infologLength);
        glGetProgramInfoLog(obj, infologLength, &charsWritten, infoLog);
        printf("%s\n",infoLog);
        free(infoLog);
    }
}


int create_default_shaders_from_sources_and_create_and_link_program(
		char *vertex_shader_source, char *fragment_shader_source,
		GLuint *vertex_shader, GLuint *fragment_shader, GLuint *shader_program) {

	GLint status;

	*vertex_shader = load_shader_from_memory(GL_VERTEX_SHADER, vertex_shader_source);
	if (!*vertex_shader) {
		return 0;
	}
	*fragment_shader = load_shader_from_memory(GL_FRAGMENT_SHADER, fragment_shader_source);
	if (!*fragment_shader) {
		glDeleteShader(*vertex_shader);
		return 0;
	}

	*shader_program = glCreateProgram();
	glAttachShader(*shader_program, *vertex_shader);
	glAttachShader(*shader_program, *fragment_shader);
	glLinkProgram(*shader_program);
	glGetProgramiv(*shader_program, GL_LINK_STATUS, &status);
	if (status == GL_FALSE) {
		printf("Niepoprawna konsolidacja programu\n");
		print_program_info_log(*shader_program);
		glDeleteShader(*vertex_shader);
		glDeleteShader(*fragment_shader);
		return 0;
	}
	glValidateProgram(*shader_program);
	glGetProgramiv(*shader_program, GL_VALIDATE_STATUS, &status);
	if (status == GL_FALSE) {
		printf("Niepoprawna walidacja programu\n");
		print_program_info_log(*shader_program);
		glDeleteShader(*vertex_shader);
		glDeleteShader(*fragment_shader);
		glDeleteProgram(*shader_program);
		return 0;
	}

	glUseProgram(*shader_program);

	return 1;

}

int create_default_shaders_from_files_and_create_and_link_program(
		char *vertex_shader_filename, char *fragment_shader_filename,
		GLuint *vertex_shader, GLuint *fragment_shader, GLuint *shader_program) {

	int result = 0;
	char *vertex_shader_source = 0;
	char *fragment_shader_source = 0;

	vertex_shader_source = text_file_read(vertex_shader_filename);
	if (!vertex_shader_source) {
		return 0;
	}
	fragment_shader_source = text_file_read(fragment_shader_filename);
	if (!fragment_shader_source) {
		free(vertex_shader_source);
		return 0;
	}

	result = create_default_shaders_from_files_and_create_and_link_program(
			vertex_shader_source, fragment_shader_source,
			vertex_shader, fragment_shader, shader_program);

	free(vertex_shader_source);
	free(fragment_shader_source);

	return result;
}
